//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

/**
 * @brief Allow socket to download limit number of bytes
 *
 * This function is called by MOD_SetDownloadLimitByRate()
 * @param limit Number of bytes to send during this 0.1 second slot 
 * @return void
 */
void CEMSocket::SetDownloadLimit(const CBytes& limit){	
	// MOD BEGIN netfinity: Accumulate download limits
	downloadLimit += limit;
	downloadLimitEnable = true;	
	if(downloadLimit > 20 * limit && downloadLimit > 4500) // Allow a maximum of 2.0 sec to accumulate or 3 * MTU
		downloadLimit = min(max(20 * limit, 4500), limit);
	// MOD BEGIN netfinity

	// CPU load improvement
	if(limit > 0 && pendingOnReceive == true){
		OnReceive(0);
	}
}

/**
 * @brief Allow socket to download limit number of bytes per second
 *
 * This function is called approx. every 0.1 second by CPartFile::ProcessDL()
 * @param limit Number of bytes to send during this 0.1 second slot given in bytes per second 
 * @return void
 */
void CEMSocket::MOD_SetDownloadLimitByRate(const CBytesPerSecond& limit){	
	CMilliSeconds	timeDiff = downloadLimitLastCall.Elapsed();
	downloadLimitLastCall.SetCurrentTick();
	if (!(timeDiff >= SEC2MS(0) && timeDiff <= SEC2MS(1)))
		timeDiff = SEC2MS(1);
	SetDownloadLimit(limit * timeDiff);
}

/**
 * @brief Allow socket to download as much data as possible
 *
 * @return void
 */
void CEMSocket::DisableDownloadLimit(){
	downloadLimitEnable = false;

	// CPU load improvement
	if(pendingOnReceive == true){
		OnReceive(0);
	}
}

/**
 * @brief Allow socket to upload a limited number of bytes
 *
 * This function is called by MOD_SetUploadLimitByRate()
 * @param limit Number of bytes to send during this 0.1 second slot 
 * @return void
 */
void CEMSocket::MOD_SetUploadLimit(const CBytes& limit){	
	// MOD BEGIN netfinity: Accumulate download limits
	uploadLimit += limit;	
	uploadLimitEnable = true;	
	if(uploadLimit > 20 * limit && uploadLimit > 4500) // Allow a maximum of 2.0 sec to accumulate or 3 * MTU, but not less than limit
		uploadLimit = min(max(20 * limit, 4500), limit); 
	// MOD BEGIN netfinity
}

/**
 * @brief Allow socket to upload limit number of bytes per second
 *
 * This function is called approx. every 0.1 second by CUploadQueue::Process()
 * @param limit Number of bytes to send during this 0.1 second slot given in bytes per second 
 * @return void
 */
void CEMSocket::MOD_SetUploadLimitByRate(const CBytesPerSecond& limit){	
	CMilliSeconds	timeDiff = uploadLimitLastCall.Elapsed();
	uploadLimitLastCall.SetCurrentTick();
	if (!(timeDiff >= SEC2MS(0) && timeDiff <= SEC2MS(1)))
		timeDiff = SEC2MS(1);
	MOD_SetUploadLimit(limit * timeDiff);
}

/**
 * @brief Allow socket to upload as much data as possible
 *
 * @return void
 */
void CEMSocket::MOD_DisableUploadLimit(){
	uploadLimitEnable = false;
}
